@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# PyCuda documentation build configuration file, created by
+# PyCUDA documentation build configuration file, created by
 # sphinx-quickstart on Fri Jun 13 00:51:19 2008.
 #
 # This file is execfile()d with the current directory set to its containing dir.
@@ -35,20 +35,18 @@ source_suffix = '.rst'
 master_doc = 'index'
 
 # General substitutions.
-project = 'PyCuda'
+project = 'PyCUDA'
 copyright = '2008, Andreas Kloeckner'
 
 # The default replacements for |version| and |release|, also used in various
 # other places throughout the built documents.
 #
 # The short X.Y version.
-import re
-ver_re = re.compile(r'version\s*=\s*"([0-9a-z.]+)"')
-version = [ver_re.search(line).group(1) 
-        for line in open("../../setup.py").readlines() 
-        if ver_re.search(line)][0]
+ver_dic = {}
+execfile("../../pycuda/__init__.py", ver_dic)
+version = ".".join(str(x) for x in ver_dic["VERSION"])
 # The full version, including alpha/beta/rc tags.
-release = version
+release = ver_dic["VERSION_TEXT"]
 
 # There are two options for replacing |today|: either, you set today to some
 # non-false value, then it is used:
@@ -144,7 +142,7 @@ htmlhelp_basename = 'PyCudadoc'
 # Grouping the document tree into LaTeX files. List of tuples
 # (source start file, target name, title, author, document class [howto/manual]).
 latex_documents = [
-  ('index', 'PyCuda.tex', 'PyCuda Documentation', 'Andreas Kloeckner', 'manual'),
+  ('index', 'pycdua.tex', 'PyCUDA Documentation', 'Andreas Kloeckner', 'manual'),
 ]
 
 # The name of an image file (relative to this directory) to place at the top of
