@@ -22,7 +22,7 @@ mod = SourceModule("""
     """)
 
 func = mod.get_function("doublify")
-func(a_gpu, block=(4,4,1), grid=(1,1))
+func(a_gpu, block=(4,4,1), grid=(1,1), shared=0)
 
 a_doubled = numpy.empty_like(a)
 cuda.memcpy_dtoh(a_doubled, a_gpu)
