@@ -76,13 +76,13 @@ def compile_plain(source, options, keep, nvcc, cache_dir):
         if '#include' in source:
             checksum.update(preprocess_source(source, options, nvcc))
         else:
-            checksum.update(source)
+            checksum.update(source.encode("utf-8"))
 
         for option in options: 
-            checksum.update(option)
+            checksum.update(option.encode("utf-8"))
         checksum.update(get_nvcc_version(nvcc))
         from pycuda.characterize import platform_bits
-        checksum.update(str(platform_bits()))
+        checksum.update(str(platform_bits()).encode("utf-8"))
 
         cache_file = checksum.hexdigest()
         cache_path = join(cache_dir, cache_file + ".cubin")
@@ -130,12 +130,11 @@ def compile_plain(source, options, keep, nvcc, cache_dir):
 
     if stdout or stderr:
         lcase_err_text = (stdout+stderr).lower()
-        if "demoted" in lcase_err_text or "demoting" in lcase_err_text:
-            from pycuda.driver import CompileError
-            raise CompileError("nvcc said it demoted types in source code it "
-                "compiled--this is likely not what you want.",
-                cmdline, stdout=stdout, stderr=stderr)
         from warnings import warn
+        if "demoted" in lcase_err_text or "demoting" in lcase_err_text:
+            warn("nvcc said it demoted types in source code it "
+                "compiled--this is likely not what you want.",
+                stacklevel=4)
         warn("The CUDA compiler suceeded, but said the following:\n"
                 +stdout+stderr, stacklevel=4)
 
@@ -287,6 +286,7 @@ class SourceModule(object):
 
         self.get_global = self.module.get_global
         self.get_texref = self.module.get_texref
+        self.get_surfref = self.module.get_surfref
 
     def _check_arch(self, arch):
         if arch is None: return
