@@ -31,14 +31,15 @@ OTHER DEALINGS IN THE SOFTWARE.
 
 
 from pycuda.tools import context_dependent_memoize
-import numpy
+import numpy as np
 from pycuda.tools import dtype_to_ctype, VectorArg, ScalarArg
+from pytools import memoize_method
 
 
 
 
 def get_elwise_module(arguments, operation,
-        name="kernel", keep=False, options=[],
+        name="kernel", keep=False, options=None,
         preamble="", loop_prep="", after_loop=""):
     from pycuda.compiler import SourceModule
     return SourceModule("""
@@ -73,25 +74,90 @@ def get_elwise_module(arguments, operation,
             },
         options=options, keep=keep)
 
+
+
+
+def get_elwise_range_module(arguments, operation,
+        name="kernel", keep=False, options=None,
+        preamble="", loop_prep="", after_loop=""):
+    from pycuda.compiler import SourceModule
+    return SourceModule("""
+        #include <pycuda-complex.hpp>
+
+        %(preamble)s
+
+        __global__ void %(name)s(%(arguments)s)
+        {
+          unsigned tid = threadIdx.x;
+          unsigned total_threads = gridDim.x*blockDim.x;
+          unsigned cta_start = blockDim.x*blockIdx.x;
+          long i;
+
+          %(loop_prep)s;
+
+          if (step < 0)
+          {
+            for (i = start + (cta_start + tid)*step;
+              i > stop; i += total_threads*step)
+            {
+              %(operation)s;
+            }
+          }
+          else
+          {
+            for (i = start + (cta_start + tid)*step; 
+              i < stop; i += total_threads*step)
+            {
+              %(operation)s;
+            }
+          }
+
+          %(after_loop)s;
+        }
+        """ % {
+            "arguments": ", ".join(arg.declarator() for arg in arguments),
+            "operation": operation,
+            "name": name,
+            "preamble": preamble,
+            "loop_prep": loop_prep,
+            "after_loop": after_loop,
+            },
+        options=options, keep=keep)
+
+
+
+
 def get_elwise_kernel_and_types(arguments, operation,
-        name="kernel", keep=False, options=[], **kwargs):
+        name="kernel", keep=False, options=None, use_range=False, **kwargs):
     if isinstance(arguments, str):
         from pycuda.tools import parse_c_arg
         arguments = [parse_c_arg(arg) for arg in arguments.split(",")]
 
-    arguments.append(ScalarArg(numpy.uintp, "n"))
+    if use_range:
+        arguments.extend([
+            ScalarArg(np.intp, "start"),
+            ScalarArg(np.intp, "stop"),
+            ScalarArg(np.intp, "step"),
+            ])
+    else:
+        arguments.append(ScalarArg(np.uintp, "n"))
 
-    mod = get_elwise_module(arguments, operation, name,
+    if use_range:
+        module_builder = get_elwise_range_module
+    else:
+        module_builder = get_elwise_module
+
+    mod = module_builder(arguments, operation, name,
             keep, options, **kwargs)
 
     from pycuda.tools import get_arg_type
     func = mod.get_function(name)
-    func.prepare("".join(arg.struct_char for arg in arguments), (1,1,1))
+    func.prepare("".join(arg.struct_char for arg in arguments))
 
     return func, arguments
 
 def get_elwise_kernel(arguments, operation,
-        name="kernel", keep=False, options=[], **kwargs):
+        name="kernel", keep=False, options=None, **kwargs):
     """Return a L{pycuda.driver.Function} that performs the same scalar operation
     on one or several vectors.
     """
@@ -105,30 +171,74 @@ def get_elwise_kernel(arguments, operation,
 
 class ElementwiseKernel:
     def __init__(self, arguments, operation,
-            name="kernel", keep=False, options=[], **kwargs):
-        self.func, self.arguments = get_elwise_kernel_and_types(
-            arguments, operation, name, keep, options, **kwargs)
+            name="kernel", keep=False, options=None, **kwargs):
 
-        assert [i for i, arg in enumerate(self.arguments)
+        self.gen_kwargs = kwargs.copy()
+        self.gen_kwargs.update(dict(keep=keep, options=options, name=name,
+            operation=operation, arguments=arguments))
+
+    @memoize_method
+    def generate_stride_kernel_and_types(self, use_range):
+        knl, arguments = get_elwise_kernel_and_types(use_range=use_range, 
+                **self.gen_kwargs)
+
+        assert [i for i, arg in enumerate(arguments)
                 if isinstance(arg, VectorArg)], \
                 "ElementwiseKernel can only be used with functions that have at least one " \
                 "vector argument"
 
-    def __call__(self, *args):
+        return knl, arguments
+
+    def __call__(self, *args, **kwargs):
         vectors = []
 
+        range_ = kwargs.pop("range", None)
+        slice_ = kwargs.pop("slice", None)
+
+        if kwargs:
+            raise TypeError("invalid keyword arguments specified: "
+                    + ", ".join(kwargs.iterkeys()))
+
         invocation_args = []
-        for arg, arg_descr in zip(args, self.arguments):
+        func, arguments = self.generate_stride_kernel_and_types(
+                range_ is not None or slice_ is not None)
+
+        for arg, arg_descr in zip(args, arguments):
             if isinstance(arg_descr, VectorArg):
+                if not arg.flags.forc:
+                    raise RuntimeError("elementwise kernel cannot "
+                            "deal with non-contiguous arrays")
+
                 vectors.append(arg)
                 invocation_args.append(arg.gpudata)
             else:
                 invocation_args.append(arg)
 
         repr_vec = vectors[0]
-        invocation_args.append(repr_vec.mem_size)
-        self.func.set_block_shape(*repr_vec._block)
-        self.func.prepared_call(repr_vec._grid, *invocation_args)
+
+        if slice_ is not None:
+            if range_ is not None:
+                raise TypeError("may not specify both range and slice "
+                        "keyword arguments")
+
+            range_ = slice(*slice_.indices(repr_vec.size))
+
+        if range_ is not None:
+            invocation_args.append(range_.start)
+            invocation_args.append(range_.stop)
+            if range_.step is None:
+                invocation_args.append(1)
+            else:
+                invocation_args.append(range_.step)
+
+            from pycuda.gpuarray import splay
+            grid, block = splay(abs(range_.stop - range_.start)//range_.step)
+        else:
+            block = repr_vec._block
+            grid = repr_vec._grid
+            invocation_args.append(repr_vec.mem_size)
+
+        func.prepared_call(grid, block, *invocation_args)
 
 
 
@@ -143,7 +253,7 @@ def get_take_kernel(dtype, idx_dtype, vec_count=1):
 
     args = [VectorArg(idx_dtype, "idx")] + [
             VectorArg(dtype, "dest"+str(i))for i in range(vec_count)] + [
-            ScalarArg(numpy.intp, "n")]
+            ScalarArg(np.intp, "n")]
     preamble = "#include <pycuda-helpers.hpp>\n\n" + "\n".join(
         "texture <%s, 1, cudaReadModeElementType> tex_src%d;" % (ctx["tex_tp"], i)
         for i in range(vec_count))
@@ -156,7 +266,7 @@ def get_take_kernel(dtype, idx_dtype, vec_count=1):
     mod = get_elwise_module(args, body, "take", preamble=preamble)
     func = mod.get_function("take")
     tex_src = [mod.get_texref("tex_src%d" % i) for i in range(vec_count)]
-    func.prepare("P"+(vec_count*"P")+numpy.dtype(numpy.uintp).char, (1,1,1), texrefs=tex_src)
+    func.prepare("P"+(vec_count*"P")+np.dtype(np.uintp).char, texrefs=tex_src)
     return func, tex_src
 
 
@@ -179,7 +289,7 @@ def get_take_put_kernel(dtype, idx_dtype, with_offsets, vec_count=1):
             ] + [
             ScalarArg(idx_dtype, "offset%d" % i)
                 for i in range(vec_count) if with_offsets
-            ] + [ScalarArg(numpy.intp, "n")]
+            ] + [ScalarArg(np.intp, "n")]
 
     preamble = "#include <pycuda-helpers.hpp>\n\n" + "\n".join(
         "texture <%s, 1, cudaReadModeElementType> tex_src%d;" % (ctx["tex_tp"], i)
@@ -206,8 +316,8 @@ def get_take_put_kernel(dtype, idx_dtype, with_offsets, vec_count=1):
     func.prepare(
             "PP"+(vec_count*"P")
             +(bool(with_offsets)*vec_count*idx_dtype.char)
-            +numpy.dtype(numpy.uintp).char,
-            (1,1,1), texrefs=tex_src)
+            +np.dtype(np.uintp).char,
+            texrefs=tex_src)
     return func, tex_src
 
 
@@ -228,7 +338,7 @@ def get_put_kernel(dtype, idx_dtype, vec_count=1):
             ] + [
             VectorArg(dtype, "src%d" % i)
                 for i in range(vec_count)
-            ] + [ScalarArg(numpy.intp, "n")]
+            ] + [ScalarArg(np.intp, "n")]
 
     body = (
             "%(idx_tp)s dest_idx = gmem_dest_idx[i];\n" % ctx
@@ -236,7 +346,7 @@ def get_put_kernel(dtype, idx_dtype, vec_count=1):
                 for i in range(vec_count)))
 
     func = get_elwise_module(args, body, "put").get_function("put")
-    func.prepare("P"+(2*vec_count*"P")+numpy.dtype(numpy.uintp).char, (1,1,1))
+    func.prepare("P"+(2*vec_count*"P")+np.dtype(np.uintp).char)
     return func
 
 
@@ -285,7 +395,7 @@ def get_linear_combination_kernel(summand_descriptors,
         summands.append("a%d*x%d[i]" % (i, i))
 
     args.append(VectorArg(dtype_z, "z"))
-    args.append(ScalarArg(numpy.uintp, "n"))
+    args.append(ScalarArg(np.uintp, "n"))
 
     mod = get_elwise_module(args,
             "z[i] = " + " + ".join(summands),
@@ -296,7 +406,7 @@ def get_linear_combination_kernel(summand_descriptors,
     func = mod.get_function("linear_combination")
     tex_src = [mod.get_texref(tn) for tn in tex_names]
     func.prepare("".join(arg.struct_char for arg in args),
-            (1,1,1), texrefs=tex_src)
+            texrefs=tex_src)
 
     return func, tex_src
 
@@ -323,27 +433,16 @@ def get_axpbz_kernel(dtype):
             "axpb")
 
 @context_dependent_memoize
-def get_multiply_kernel(dtype_x, dtype_y, dtype_z):
+def get_binary_op_kernel(dtype_x, dtype_y, dtype_z, operator):
     return get_elwise_kernel(
             "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
                 "tp_x": dtype_to_ctype(dtype_x),
                 "tp_y": dtype_to_ctype(dtype_y),
                 "tp_z": dtype_to_ctype(dtype_z),
                 },
-            "z[i] = x[i] * y[i]",
+            "z[i] = x[i] %s y[i]" % operator,
             "multiply")
 
-@context_dependent_memoize
-def get_divide_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] / y[i]",
-            "divide")
-
 @context_dependent_memoize
 def get_rdivide_elwise_kernel(dtype):
     return get_elwise_kernel(
@@ -365,7 +464,7 @@ def get_binary_func_kernel(func, dtype_x, dtype_y, dtype_z):
             func+"_kernel")
 
 def get_binary_minmax_kernel(func, dtype_x, dtype_y, dtype_z):
-    if not numpy.float64 in [dtype_x, dtype_y]:
+    if not np.float64 in [dtype_x, dtype_y]:
         func = func +"f"
 
     from pytools import any
@@ -433,7 +532,7 @@ def get_arange_kernel(dtype):
 
 @context_dependent_memoize
 def get_pow_kernel(dtype):
-    if dtype == numpy.float64:
+    if dtype == np.float64:
         func = "pow"
     else:
         func = "powf"
@@ -447,7 +546,7 @@ def get_pow_kernel(dtype):
 
 @context_dependent_memoize
 def get_pow_array_kernel(dtype_x, dtype_y, dtype_z):
-    if numpy.float64 in [dtype_x, dtype_y]:
+    if np.float64 in [dtype_x, dtype_y]:
         func = "pow"
     else:
         func = "powf"
@@ -519,72 +618,3 @@ def get_if_positive_kernel(crit_dtype, dtype):
             ],
             "result[i] = crit[i] > 0 ? then_[i] : else_[i]",
             "if_positive")
-
-
-
-@context_dependent_memoize
-def get_eq_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] == y[i]",
-            "eq")        
-
-
-@context_dependent_memoize
-def get_ne_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] != y[i]",
-            "ne")      
-
-@context_dependent_memoize
-def get_le_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] <= y[i]",
-            "lt")          
-
-@context_dependent_memoize
-def get_ge_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] >= y[i]",
-            "lt") 
-
-@context_dependent_memoize
-def get_lt_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] < y[i]",
-            "lt")           
-
-@context_dependent_memoize
-def get_gt_kernel(dtype_x, dtype_y, dtype_z):
-    return get_elwise_kernel(
-            "%(tp_x)s *x, %(tp_y)s *y, %(tp_z)s *z" % {
-                "tp_x": dtype_to_ctype(dtype_x),
-                "tp_y": dtype_to_ctype(dtype_y),
-                "tp_z": dtype_to_ctype(dtype_z),
-                },
-            "z[i] = x[i] > y[i]",
-            "gt")       
