@@ -8,7 +8,14 @@ from pycuda.compyte.array import (
         f_contiguous_strides as _f_contiguous_strides, 
         c_contiguous_strides as _c_contiguous_strides, 
         ArrayFlags as _ArrayFlags,
-        get_common_dtype as _get_common_dtype)
+        get_common_dtype as _get_common_dtype_base)
+from pycuda.characterize import has_double_support
+
+
+
+
+def _get_common_dtype(obj1, obj2):
+    return _get_common_dtype_base(obj1, obj2, has_double_support())
 
 
 
@@ -199,10 +206,13 @@ class GPUArray(object):
     def set(self, ary):
         assert ary.size == self.size
         assert ary.dtype == self.dtype
+        if ary.strides != self.strides:
+            from warnings import warn
+            warn("Setting array from one with different strides/storage order. "
+                    "This will cease to work in 2013.x.",
+                    stacklevel=2)
 
         assert self.flags.forc
-        if not ary.flags.forc:
-            raise RuntimeError("cannot set from non-contiguous array")
 
         if self.size:
             drv.memcpy_htod(self.gpudata, ary)
@@ -210,6 +220,12 @@ class GPUArray(object):
     def set_async(self, ary, stream=None):
         assert ary.size == self.size
         assert ary.dtype == self.dtype
+        if ary.strides != self.strides:
+            from warnings import warn
+            warn("Setting array from one with different strides/storage order. "
+                    "This will cease to work in 2013.x.",
+                    stacklevel=2)
+
         assert self.flags.forc
 
         if not ary.flags.forc:
@@ -325,8 +341,6 @@ class GPUArray(object):
             raise RuntimeError("only contiguous arrays may "
                     "be used as arguments to this operation")
 
-        assert self.dtype == np.float32
-
         func = elementwise.get_rdivide_elwise_kernel(self.dtype)
         func.prepared_async_call(self._grid, self._block, stream,
                 self.gpudata, other,
@@ -1119,7 +1133,7 @@ def dot(a, b, dtype=None, stream=None):
 
 def subset_dot(subset, a, b, dtype=None, stream=None):
     from pycuda.reduction import get_subset_dot_kernel
-    krnl = get_subset_dot_kernel(dtype, a.dtype, b.dtype)
+    krnl = get_subset_dot_kernel(dtype, subset.dtype, a.dtype, b.dtype)
     return krnl(subset, a, b, stream=stream)
 
 def _make_minmax_kernel(what):
@@ -1139,7 +1153,7 @@ def _make_subset_minmax_kernel(what):
     def f(subset, a, stream=None):
         from pycuda.reduction import get_subset_minmax_kernel
         import pycuda.reduction
-        krnl = get_subset_minmax_kernel(what, a.dtype)
+        krnl = get_subset_minmax_kernel(what, a.dtype, subset.dtype)
         return krnl(subset, a,  stream=stream)
 
     return f
