@@ -576,6 +576,15 @@ namespace pycuda
         return m_valid;
       }
 
+      static boost::shared_ptr<context> attach(unsigned int flags)
+      {
+        CUcontext current;
+        CUDAPP_CALL_GUARDED(cuCtxAttach, (&current, flags));
+        boost::shared_ptr<context> result(new context(current));
+        context_stack::get().push(result);
+        return result;
+      }
+
       void detach()
       {
         if (m_valid)
@@ -756,6 +765,10 @@ namespace pycuda
 
 
 
+
+
+
+
 #if CUDAPP_CUDA_VERSION >= 2000
   inline
   void context_push(boost::shared_ptr<context> ctx)
