@@ -54,6 +54,13 @@ struct complex {
     _M_im = __z._M_im;
     return *this;
   }
+  __device__
+  volatile _Self& operator=(const _Self& __z) volatile {
+    _M_re = __z._M_re;
+    _M_im = __z._M_im;
+    return *this;
+  }
+
 
   template <class _Tp2>
   __device__
@@ -67,6 +74,13 @@ struct complex {
     _M_im = __z._M_im;
     return *this;
   }
+  template <class _Tp2>
+  __device__
+  volatile _Self& operator=(const complex<_Tp2>& __z) volatile {
+    _M_re = __z._M_re;
+    _M_im = __z._M_im;
+    return *this;
+  }
 
   // Element access.
   __device__
@@ -83,6 +97,12 @@ struct complex {
     return *this;
   }
   __device__
+  volatile _Self& operator= (const value_type& __x) volatile {
+    _M_re = __x;
+    _M_im = 0;
+    return *this;
+  }
+  __device__
   _Self& operator+= (const value_type& __x) {
     _M_re += __x;
     return *this;
@@ -211,6 +231,11 @@ struct _STLP_CLASS_DECLSPEC complex<float> {
     _M_im = 0.0f;
     return *this;
   }
+  volatile __device__ _Self& operator= (value_type __x) volatile {
+    _M_re = __x;
+    _M_im = 0.0f;
+    return *this;
+  }
   __device__ _Self& operator+= (value_type __x) {
     _M_re += __x;
     return *this;
@@ -248,6 +273,15 @@ struct _STLP_CLASS_DECLSPEC complex<float> {
     return *this;
   }
 
+  template <class _Tp2>
+  __device__ 
+  volatile complex<float>& operator=(const complex<_Tp2>& __z) volatile {
+    _M_re = __z._M_re;
+    _M_im = __z._M_im;
+    return *this;
+  }
+
+
   template <class _Tp2>
   __device__ 
   complex<float>& operator+= (const complex<_Tp2>& __z) {
@@ -292,6 +326,14 @@ struct _STLP_CLASS_DECLSPEC complex<float> {
     return *this;
   }
 
+  __device__ 
+  volatile _Self& operator=(const _Self& __z) volatile {
+    _M_re = __z._M_re;
+    _M_im = __z._M_im;
+    return *this;
+  }
+
+
   __device__ 
   _Self& operator+= (const _Self& __z) {
     _M_re += __z._M_re;
@@ -361,6 +403,12 @@ struct _STLP_CLASS_DECLSPEC complex<double> {
     return *this;
   }
   __device__
+  volatile _Self& operator= (value_type __x) volatile {
+    _M_re = __x;
+    _M_im = 0.0;
+    return *this;
+  }
+  __device__
   _Self& operator+= (value_type __x) {
     _M_re += __x;
     return *this;
@@ -401,6 +449,14 @@ struct _STLP_CLASS_DECLSPEC complex<double> {
     return *this;
   }
 
+  template <class _Tp2>
+  __device__
+  volatile complex<double>& operator=(const volatile complex<_Tp2>& __z) {
+    _M_re = __z._M_re;
+    _M_im = __z._M_im;
+    return *this;
+  }
+
   template <class _Tp2>
   __device__
   complex<double>& operator+= (const complex<_Tp2>& __z) {
@@ -446,6 +502,13 @@ struct _STLP_CLASS_DECLSPEC complex<double> {
     return *this;
   }
 
+  __device__ 
+  volatile _Self& operator=(const _Self& __z) volatile {
+    _M_re = __z._M_re;
+    _M_im = __z._M_im;
+    return *this;
+  }
+
   __device__ 
   _Self& operator+= (const _Self& __z) {
     _M_re += __z._M_re;
@@ -557,6 +620,13 @@ inline complex<_Tp>
 operator+(const complex<_Tp>& __z1, const complex<_Tp>& __z2)
 { return complex<_Tp>(__z1._M_re + __z2._M_re, __z1._M_im + __z2._M_im); }
 
+template <class _Tp>
+__device__
+inline complex<_Tp>
+operator+(const volatile complex<_Tp>& __z1, const volatile complex<_Tp>& __z2)
+{ return complex<_Tp>(__z1._M_re + __z2._M_re, __z1._M_im + __z2._M_im); }
+
+
 template <class _Tp>
 __device__
 inline complex<_Tp> __device__
@@ -571,6 +641,14 @@ operator*(const complex<_Tp>& __z1, const complex<_Tp>& __z2) {
                       __z1._M_re * __z2._M_im + __z1._M_im * __z2._M_re);
 }
 
+template <class _Tp>
+__device__
+inline complex<_Tp> __device__
+operator*(const volatile complex<_Tp>& __z1, const volatile complex<_Tp>& __z2) {
+  return complex<_Tp>(__z1._M_re * __z2._M_re - __z1._M_im * __z2._M_im,
+                      __z1._M_re * __z2._M_im + __z1._M_im * __z2._M_re);
+}
+
 template <class _Tp>
 __device__
 inline complex<_Tp> __device__
@@ -699,7 +777,8 @@ __device__ complex<float> exp(const complex<float>&);
 __device__ complex<float>  log(const complex<float>&);
 __device__ complex<float> log10(const complex<float>&);
 
-__device__ complex<float> pow(const complex<float>&, int);
+// uses some stlport-private power thing
+// __device__ complex<float> pow(const complex<float>&, int);
 __device__ complex<float> pow(const complex<float>&, const float&);
 __device__ complex<float> pow(const float&, const complex<float>&);
 __device__ complex<float> pow(const complex<float>&, const complex<float>&);
@@ -718,7 +797,8 @@ __device__ complex<double> exp(const complex<double>&);
 __device__ complex<double> log(const complex<double>&);
 __device__ complex<double> log10(const complex<double>&);
 
-__device__ complex<double> pow(const complex<double>&, int);
+// uses some stlport-private power thing
+// __device__ complex<double> pow(const complex<double>&, int);
 __device__ complex<double> pow(const complex<double>&, const double&);
 __device__ complex<double> pow(const double&, const complex<double>&);
 __device__ complex<double> pow(const complex<double>&, const complex<double>&);
