@@ -896,6 +896,8 @@ BOOST_PYTHON_MODULE(_driver)
       .def(py::self != py::self)
       .def("__hash__", &cl::hash)
 
+      .def("attach", &cl::attach, (py::arg("flags")=0))
+      .staticmethod("attach")
       .DEF_SIMPLE_METHOD(detach)
 
 #if CUDAPP_CUDA_VERSION >= 2000
@@ -1005,7 +1007,7 @@ BOOST_PYTHON_MODULE(_driver)
       .DEF_SIMPLE_METHOD(set_cache_config)
 #endif
 #if CUDAPP_CUDA_VERSION >= 4000
-      .def("launch_kernel", &cl::launch_kernel)
+      .def("_launch_kernel", &cl::launch_kernel)
 #endif
       ;
   }
@@ -1351,7 +1353,8 @@ BOOST_PYTHON_MODULE(_driver)
   {
     typedef surface_reference cl;
     py::class_<cl, boost::noncopyable>("SurfaceReference", py::no_init)
-      .DEF_SIMPLE_METHOD(set_array)
+      .def("set_array", &cl::set_array,
+          (py::arg("array"), py::arg("flags")=0))
       .def("get_array", &cl::get_array,
           py::return_value_policy<py::manage_new_object>())
       ;
