@@ -555,16 +555,18 @@ class TestGPUArray:
     @mark_cuda_test
     def test_dot(self):
         from pycuda.curandom import rand as curand
-        a_gpu = curand((200000,))
-        a = a_gpu.get()
-        b_gpu = curand((200000,))
-        b = b_gpu.get()
+        for l in [2,3,4,5,6,7, 31, 32, 33, 127, 128, 129, 255, 256, 257, 16384 - 993,
+                20000]:
+            a_gpu = curand((l,))
+            a = a_gpu.get()
+            b_gpu = curand((l,))
+            b = b_gpu.get()
 
-        dot_ab = np.dot(a, b)
+            dot_ab = np.dot(a, b)
 
-        dot_ab_gpu = gpuarray.dot(a_gpu, b_gpu).get()
+            dot_ab_gpu = gpuarray.dot(a_gpu, b_gpu).get()
 
-        assert abs(dot_ab_gpu-dot_ab)/abs(dot_ab) < 1e-4
+            assert abs(dot_ab_gpu-dot_ab)/abs(dot_ab) < 1e-4
 
     @mark_cuda_test
     def test_slice(self):
@@ -599,8 +601,8 @@ class TestGPUArray:
         max_a_b_gpu = gpuarray.maximum(a_gpu, b_gpu)
         min_a_b_gpu = gpuarray.minimum(a_gpu, b_gpu)
 
-        print max_a_b_gpu
-        print np.maximum(a, b)
+        print (max_a_b_gpu)
+        print (np.maximum(a, b))
 
         assert la.norm(max_a_b_gpu.get()- np.maximum(a, b)) == 0
         assert la.norm(min_a_b_gpu.get()- np.minimum(a, b)) == 0
@@ -717,9 +719,9 @@ class TestGPUArray:
     def test_stride_preservation(self):
         A = np.random.rand(3,3)
         AT = A.T
-        print AT.flags.f_contiguous, AT.flags.c_contiguous
+        print (AT.flags.f_contiguous, AT.flags.c_contiguous)
         AT_GPU = gpuarray.to_gpu(AT)
-        print AT_GPU.flags.f_contiguous, AT_GPU.flags.c_contiguous
+        print (AT_GPU.flags.f_contiguous, AT_GPU.flags.c_contiguous)
         assert np.allclose(AT_GPU.get(),AT)
 
     @mark_cuda_test
@@ -836,7 +838,7 @@ if __name__ == "__main__":
 
     import sys
     if len(sys.argv) > 1:
-        exec sys.argv[1]
+        exec (sys.argv[1])
     else:
         from py.test.cmdline import main
         main([__file__])
